#include "Python.h"
#include "structmember.h"
#include "windows.h"
#include "support.h"

/* Two dictionaries to implement WNDPROC as Python functions:
 *
 * wnd_to_func :- find Python function by window handle
 * wnd_funcs   :- known Python WNDPROC functions
 * name_to_func :- find Python function by window class name
 */

static PyObject *wnd_to_func;
static PyObject *wnd_funcs;
static PyObject *name_to_func;

static void check_dicts()
{
    if (wnd_to_func != NULL)
	return;
    wnd_to_func = PyDict_New();
    wnd_funcs = PyDict_New();
    name_to_func = PyDict_New();
}

static int add_known_func(PyObject *func)
{
    PyObject *key, *count;
    int ok;

    key = PyInt_FromLong((int)func);
    if (key == NULL)
	return 0;

    count = PyDict_GetItem(wnd_to_func, key); /* borrowed */
    if (count != NULL)
	count = PyInt_FromLong(PyInt_AsLong(count) + 1);
    else
	count = PyInt_FromLong(1);
    if (count == NULL) {
	Py_DECREF(key);
	return 0;
    }

    ok = PyDict_SetItem(wnd_funcs, key, count) == 0;
    Py_DECREF(key);
    Py_DECREF(count);

    return 1;
}

static int del_known_func(PyObject *func)
{
    PyObject *key, *count;
    int num, ok;

    key = PyInt_FromLong((int)func);
    if (key == NULL)
	return 0;

    count = PyDict_GetItem(wnd_to_func, key); /* borrowed */
    if (count == NULL) {
	Py_DECREF(key);
	return 0;
    }
    num = PyInt_AsLong(count) - 1;
    if (num == 0) {
	ok = PyDict_DelItem(wnd_to_func, key);
	Py_DECREF(key);
	return ok;
    }

    count = PyInt_FromLong(num);
    if (count == NULL) {
	Py_DECREF(key);
	return 0;
    }
    ok = PyDict_SetItem(wnd_funcs, key, count) == 0;
    Py_XDECREF(key);
    Py_XDECREF(count);
    return ok;
}

static int is_known_func(int func)
{
    PyObject *key, *count;

    key = PyInt_FromLong(func);
    if (key == NULL)
        return 0;

    count = PyDict_GetItem(wnd_funcs, key); /* borrowed */
    Py_DECREF(key);

    return count != NULL;
}

static int wnd_func_add(HWND wnd, PyObject *func, PyObject **old_func)
{
    int ok;
    PyObject *key;

    key = PyInt_FromLong((int)wnd);
    if (key == NULL)
	return 0;

    *old_func = PyDict_GetItem(wnd_to_func, key);
    if (*old_func != NULL)
	del_known_func(*old_func);

    ok = PyDict_SetItem(wnd_to_func, key, func) == 0;
    Py_DECREF(key);

    return add_known_func(func);
}

static PyObject *wnd_func_find(HWND wnd)
{
    PyObject *key;

    key = PyInt_FromLong((int)wnd);
    if (key == NULL)
        return NULL;

    return PyDict_GetItem(wnd_to_func, key);
}

static int wnd_func_remove(HWND wnd)
{
    PyObject *key, *func;
    int ok;

    key = PyInt_FromLong((int)wnd);
    if (key == NULL)
        return 0;

    func = PyDict_GetItem(wnd_to_func, key);
    if (func != NULL)
	del_known_func(func);

    ok = PyDict_DelItem(wnd_to_func, key);
    Py_DECREF(key);

    return ok;
}

/* Bounce messages up into the correct Python function
 */
LRESULT CALLBACK dlg_proc_trampoline(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    PyObject *func, *old_func;
    BOOL result = 0;

    check_dicts();

    if (msg == WM_INITDIALOG) {
	func = (PyObject *)lparam;
	if (!wnd_func_add(wnd, func, &old_func))
	    func = NULL;
	lparam = 0;
    } else
	func = wnd_func_find(wnd);

    if (func != NULL) {
	PyObject *result_obj = NULL;

	result_obj = PyObject_CallFunction(func, "iiii", wnd, msg,
					   wparam, lparam);
	if (result_obj != NULL && PyInt_Check(result_obj))
	    result = PyInt_AsLong(result_obj);

	Py_XDECREF(result_obj);
	if (result_obj == NULL)
	    EndDialog(wnd, -1);

	if (msg == WM_DESTROY)
	    wnd_func_remove(wnd);
    }

    return result;
}

LRESULT CALLBACK wnd_proc_trampoline(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    PyObject *func;
    BOOL result = 0;

    check_dicts();

    func = wnd_func_find(wnd);

    if (func != NULL) {
	PyObject *result_obj = NULL;

	result_obj = PyObject_CallFunction(func, "iiii", wnd, msg,
					   wparam, lparam);
	if (result_obj != NULL && PyInt_Check(result_obj))
	    result = PyInt_AsLong(result_obj);
	Py_XDECREF(result_obj);

	if (msg == WM_DESTROY)
	    wnd_func_remove(wnd);
    }

    return result;
}

LRESULT wnd_proc_call(int wnd_proc, int wnd, int msg, int wparam, int lparam)
{
    check_dicts();

    if (is_known_func(wnd_proc)) {
	PyObject *result_obj = NULL;
	LRESULT result;

	result_obj = PyObject_CallFunction((PyObject*)wnd_proc, "iiii",
					   wnd, msg, wparam, lparam);
	if (result_obj != NULL && PyInt_Check(result_obj))
	    result = PyInt_AsLong(result_obj);
	Py_XDECREF(result_obj);
	return result;
    } else
	return CallWindowProc((WNDPROC)wnd_proc, (HWND)wnd, (UINT)msg,
			      (WPARAM)wparam, (LPARAM)lparam);
}

/* If index is GWL_WNDPROC then new_long can be a callable, otherwise
 * it must be an integer.
 */
PyObject *wnd_proc_setwindowlong(int wnd, int index, PyObject *new_long)
{
    LONG old_long;
    PyObject *old_func;

    check_dicts();

    switch (index) {
    default:
	if (!PyInt_Check(new_long)) {
	    PyErr_SetString(PyExc_ValueError, "new_long must be integer");
	    return NULL;
	}
	old_long = SetWindowLong((HWND)wnd, index,
				 (LONG)PyInt_AsLong(new_long));
	return PyInt_FromLong(old_long);
    case GWL_WNDPROC:
    case DWL_DLGPROC:
	/* We could either restore an integer WNDPROC or install a
	 * callable Python function.
	 */
	if (PyInt_Check(new_long)) {
	    old_long = SetWindowLong((HWND)wnd, index,
				     (LONG)PyInt_AsLong(new_long));
	    return PyInt_FromLong(old_long);
	}
	/* Installing Python WNDPROC
	 */
	if (!PyCallable_Check(new_long)) {
	    PyErr_SetString(PyExc_ValueError, "new_long must be integer or callable");
            return NULL;
	}
	wnd_func_add((HWND)wnd, new_long, &old_func);
	old_long = SetWindowLong((HWND)wnd, index, (LONG)wnd_proc_trampoline);
	if (old_long == (LONG)wnd_proc_trampoline) {
	    if (old_func == NULL) {
		fprintf(stderr, "old_func should have been non-NULL\n");
		return PyInt_FromLong(0);
	    }
	    Py_INCREF(old_func);
	    return old_func;
	}
	return PyInt_FromLong(old_long);
    }
}

PyObject *wnd_proc_getwindowlong(int wnd, int index)
{
    LONG result;

    check_dicts();

    switch (index) {
    default:
	return PyInt_FromLong(GetWindowLong((HWND)wnd, index));

    case GWL_WNDPROC:
    case DWL_DLGPROC:
	result = GetWindowLong((HWND)wnd, index);
	if (result == (LONG)dlg_proc_trampoline
	    || result == (LONG)wnd_proc_trampoline) {
	    PyObject *func;

	    func = wnd_func_find((HWND)wnd);
	    Py_INCREF(func);
	    return func;
	}
	return PyInt_FromLong(result);
    }
}

/* Window Class Procedure
 */
int name_proc_add(const char *name, PyObject *func)
{
    check_dicts();
    return PyDict_SetItemString(name_to_func, (char*)name, func) == 0;
}

static PyObject *name_func_find(char *name)
{
    return PyDict_GetItemString(name_to_func, name);
}

static int name_func_remove(char *name)
{
    return PyDict_DelItemString(name_to_func, name) == 0;
}

/* Bounce window messages up into the correct Python function
 */
LRESULT CALLBACK name_proc_trampoline(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    PyObject *func, *old_func;
    int num;
    char name[256];
    LRESULT result = 0;

    check_dicts();

    num = GetClassName(wnd, name, sizeof(name));
    func = name_func_find(name);

    if (func != NULL) {
	PyObject *result_obj = NULL;

	wnd_func_add(wnd, func, &old_func);
	SetWindowLong(wnd, GWL_WNDPROC, (LONG)wnd_proc_trampoline);

	result_obj = PyObject_CallFunction(func, "iiii", wnd, msg,
					   wparam, lparam);
	if (result_obj != NULL && PyInt_Check(result_obj))
	    result = PyInt_AsLong(result_obj);

	Py_XDECREF(result_obj);
    }

    return result;
}

LONG name_proc_setclasslong(int wnd, int index, int new_long)
{
    check_dicts();

    return SetClassLong((HWND)wnd, index, (LONG)new_long);
}

LONG name_proc_getclasslong(int wnd, int index)
{
    check_dicts();

    return GetClassLong((HWND)wnd, index);
}

typedef struct TIMERPROCFunc {
    struct TIMERPROCFunc *next;
    HWND wnd;
    UINT timer_id;
    PyObject *func;
} TIMERPROCFunc;

static TIMERPROCFunc *timerproc_funcs;

int timerproc_add(HWND wnd, UINT timer_id, PyObject *func)
{
    TIMERPROCFunc *wf;

    if (func == Py_None)
	return 0;

    wf = malloc(sizeof(*wf));
    if (wf == NULL) {
	PyErr_NoMemory();
	return -1;
    }
    wf->wnd = wnd;
    wf->timer_id = timer_id;
    Py_INCREF(func);
    wf->func = func;
    wf->next = timerproc_funcs;
    timerproc_funcs = wf;

    return 0;
}

void timerproc_remove(HWND wnd, UINT timer_id)
{
    TIMERPROCFunc *wf, *prev;

    for (prev = NULL, wf = timerproc_funcs; wf; prev = wf, wf = wf->next)
	if (wf->wnd == wnd && wf->timer_id == timer_id) {
	    if (prev != NULL)
		prev->next = wf->next;
	    else
		timerproc_funcs = wf->next;
	    Py_XDECREF(wf->func);
	    free(wf);
	    return;
	}
}

static TIMERPROCFunc *timerproc_func_find(HWND wnd, UINT timer_id)
{
    TIMERPROCFunc *wf;

    for (wf = timerproc_funcs; wf; wf = wf->next)
	if (wf->wnd == wnd && wf->timer_id == timer_id)
	    return wf;

    return NULL;
}

void CALLBACK timerproc_trampoline(HWND wnd, UINT msg, UINT event, DWORD time)
{
    TIMERPROCFunc *wf;

    wf = timerproc_func_find(wnd, event);
    if (wf != NULL) {
	PyObject *result_obj;

	result_obj = PyObject_CallFunction(wf->func, "iiii", wnd, msg,
					   event, time);
	Py_XDECREF(result_obj);
    }
}

int proxy_setattr_callable(char *name, PyObject *v, PyObject **obj)
{
    if (v == Py_None) {
	Py_XDECREF(*obj);
	*obj = NULL;
    } else if (!PyCallable_Check(v)) {
	char msg[256];
	sprintf(msg, "%s not callable", name);
	PyErr_SetString(PyExc_ValueError, msg);
	return -1;
    } else {
	Py_XDECREF(*obj);
	*obj = v;
	Py_INCREF(*obj);
    }
    return 0;
}

int proxy_setattr_string(char *name, PyObject *v, PyObject **obj, char **str)
{
    if (v == Py_None) {
	Py_XDECREF(*obj);
	*obj = NULL;
	*str = NULL;
    } else if (PyString_Check(v)) {
	Py_XDECREF(*obj);
	*obj = v;
	Py_INCREF(*obj);
	*str = PyString_AsString(v);
    } else if (PyInt_Check(v)) {
	Py_XDECREF(*obj);
	*obj = v;
	Py_INCREF(*obj);
	*str = (char*)PyInt_AsLong(v);
    } else {
	char msg[256];
	sprintf(msg, "%s not string or int", name);
	PyErr_SetString(PyExc_ValueError, name);
	return -1;
    }
    return 0;
}

PyObject *proxy_getattr(PyObject *obj)
{
    if (obj == NULL)
	return Py_BuildValue("");
    Py_INCREF(obj);
    return obj;
}

int check_nullcallable(char *name, PyObject *obj)
{
    char msg[256];

    if (obj == Py_None || PyCallable_Check(obj))
	return 1;

    sprintf(msg, "%s must be callable or None", name);
    PyErr_SetString(PyExc_ValueError, msg);
    return 0;
}

int conv_nullstruct(char *name, PyTypeObject *type, PyObject *obj, void **arg, int offset)
{
    char msg[128];

    if (obj == Py_None) {
	*arg = NULL;
	return 1;
    }
    if (obj->ob_type == type) {
	*arg = ((char *)obj) + offset;
	return 1;
    }

    sprintf(msg, "%s must be %s or None", name, type->tp_name);
    PyErr_SetString(PyExc_ValueError, msg);
    return 0;
}

static void Buffer_dealloc(Buffer_Obj *self)
{
    if (self->buff)
	PyMem_DEL(self->buff);
    PyMem_DEL(self);
}

static char Buffer__doc__[] = "Opaque Buffer";

PyTypeObject Buffer_Type = {
    PyObject_HEAD_INIT(0)
    0,                          /*ob_size*/
    "Buffer",                   /*tp_name*/
    sizeof(Buffer_Obj),         /*tp_basicsize*/
    0,                          /*tp_itemsize*/
    /* methods */
    (destructor)Buffer_dealloc, /*tp_dealloc*/
    (printfunc)0,               /*tp_print*/
    (getattrfunc)0,             /*tp_getattr*/
    (setattrfunc)0,             /*tp_setattr*/
    (cmpfunc)0,                 /*tp_compare*/
    (reprfunc)0,                /*tp_repr*/
    0,                          /*tp_as_number*/
    0,                          /*tp_as_sequence*/
    0,                          /*tp_as_mapping*/
    (hashfunc)0,                /*tp_hash*/
    (ternaryfunc)0,             /*tp_call*/
    (reprfunc)0,                /*tp_str*/

    /* Space for future expansion */
    0L, 0L, 0L, 0L,
    Buffer__doc__                /* Documentation string */
};

int buffer_ensure_size(Buffer_Obj *self, int size)
{
    if (self->size < size) {
	if (self->buff)
	    PyMem_DEL(self->buff);
	self->buff = PyMem_Malloc(size);
	if (self->buff == NULL)
	    return 0;
	self->size = size;
    }
    return 1;
}

PyObject *Buffer()
{
    Buffer_Obj *self;

    self = PyObject_NEW(Buffer_Obj, &Buffer_Type);
    if (self == NULL)
        return NULL;

    self->buff = NULL;
    self->len = self->size = 0;

    return (PyObject *)self;
}

int conv_nullstr(char *name, PyObject *obj, char **arg)
{
    if (obj == Py_None)
	*arg = NULL;
    else if (PyString_Check(obj))
	*arg = PyString_AsString(obj);
    else {
	char msg[128];

	sprintf(msg, "%s must be string or None", name);
        PyErr_SetString(PyExc_ValueError, msg);
	return 0;
    }
    return 1;
}

int conv_intstr(char *name, PyObject *obj, char **arg)
{
    if (PyInt_Check(obj))
	*arg = (char*)PyInt_AsLong(obj);
    else if (PyString_Check(obj))
	*arg = PyString_AsString(obj);
    else {
	char msg[128];

	sprintf(msg, "%s must be string or int", name);
        PyErr_SetString(PyExc_ValueError, msg);
	return 0;
    }
    return 1;
}

int conv_intnullstr(char *name, PyObject *obj, char **arg)
{
    if (obj == Py_None)
	*arg = NULL;
    else if (PyInt_Check(obj))
	*arg = (char*)PyInt_AsLong(obj);
    else if (PyString_Check(obj))
	*arg = PyString_AsString(obj);
    else {
	char msg[128];

	sprintf(msg, "%s must be string, int or None", name);
        PyErr_SetString(PyExc_ValueError, msg);
	return 0;
    }
    return 1;
}

static void HOSTENT_dealloc(HOSTENT_Obj *self)
{
    Py_XDECREF(self->h_name);
    Py_XDECREF(self->h_aliases);
    Py_XDECREF(self->h_addr_list);
    PyMem_DEL(self);
}

static struct memberlist HOSTENT_memberlist[] = {
    { "h_name",      T_OBJECT, offsetof(HOSTENT_Obj, h_name),      RO },
    { "h_aliases",   T_OBJECT, offsetof(HOSTENT_Obj, h_aliases),   RO },
    { "h_addrtype",  T_INT,    offsetof(HOSTENT_Obj, h_addrtype),  RO },
    { "h_length",    T_INT,    offsetof(HOSTENT_Obj, h_length),    RO },
    { "h_addr_list", T_OBJECT, offsetof(HOSTENT_Obj, h_addr_list), RO },
    { NULL }
};

static PyObject *HOSTENT_getattr(HOSTENT_Obj *self, char *name)
{
    return PyMember_Get((char *)self, HOSTENT_memberlist, name);
}

static int HOSTENT_setattr(HOSTENT_Obj *self, char *name, PyObject *v)
{
    return PyMember_Set((char *)self, HOSTENT_memberlist, name, v);
}

static char HOSTENT__doc__[] = "Wrapper for HOSTENT structure";

PyTypeObject HOSTENT_Type = {
    PyObject_HEAD_INIT(0)
    0,                          /*ob_size*/
    "HOSTENT",                  /*tp_name*/
    sizeof(HOSTENT_Obj),        /*tp_basicsize*/
    0,                          /*tp_itemsize*/
    /* methods */
    (destructor)HOSTENT_dealloc,/*tp_dealloc*/
    (printfunc)0,               /*tp_print*/
    (getattrfunc)HOSTENT_getattr, /*tp_getattr*/
    (setattrfunc)HOSTENT_setattr, /*tp_setattr*/
    (cmpfunc)0,                 /*tp_compare*/
    (reprfunc)0,                /*tp_repr*/
    0,                          /*tp_as_number*/
    0,                          /*tp_as_sequence*/
    0,                          /*tp_as_mapping*/
    (hashfunc)0,                /*tp_hash*/
    (ternaryfunc)0,             /*tp_call*/
    (reprfunc)0,                /*tp_str*/

    /* Space for future expansion */
    0L, 0L, 0L, 0L,
    HOSTENT__doc__                /* Documentation string */
};

PyObject *HOSTENT_new(HOSTENT *ent)
{
    HOSTENT_Obj *self;
    PyObject *obj;
    int num;

    self = PyObject_NEW(HOSTENT_Obj, &HOSTENT_Type);
    if (self == NULL)
        return NULL;

    obj = NULL;
    self->h_name = self->h_aliases = self->h_addr_list = NULL;

    self->h_name = PyString_FromString(ent->h_name);
    if (self->h_name == NULL)
	goto error;

    for (num = 0; ent->h_aliases[num] != NULL; num++)
	;
    self->h_aliases = PyTuple_New(num);
    if (self->h_aliases == NULL)
	goto error;
    for (num = 0; ent->h_aliases[num] != NULL; num++) {
	obj = PyString_FromString(ent->h_aliases[num]);
	if (obj == NULL)
	    goto error;
	if (PyTuple_SetItem(self->h_aliases, num, obj))
	    goto error;
	obj = NULL;		/* stolen */
    }

    self->h_addrtype = ent->h_addrtype;
    self->h_length = ent->h_length;

    for (num = 0; ent->h_addr_list[num] != NULL; num++)
	;
    self->h_addr_list = PyTuple_New(num);
    if (self->h_addr_list == NULL)
	goto error;
    for (num = 0; ent->h_addr_list[num] != NULL; num++) {
	obj = PyInt_FromLong(*(int*)ent->h_addr_list[num]);
	if (obj == NULL)
	    goto error;
	if (PyTuple_SetItem(self->h_addr_list, num, obj))
	    goto error;
	obj = NULL;		/* stolen */
    }

    return (PyObject *)self;

error:
    Py_XDECREF(self);
    Py_XDECREF(obj);
    return NULL;
}

PyObject *hostent_from_buffer(Buffer_Obj *buff)
{
    if (buff->size <= sizeof(HOSTENT)) {
	PyErr_SetString(PyExc_ValueError, "buffer too small");
	return NULL;
    }

    return HOSTENT_new((HOSTENT *)buff->buff);
}

static void SERVENT_dealloc(SERVENT_Obj *self)
{
    Py_XDECREF(self->s_name);
    Py_XDECREF(self->s_aliases);
    Py_XDECREF(self->s_proto);
    PyMem_DEL(self);
}

static struct memberlist SERVENT_memberlist[] = {
    { "s_name",    T_OBJECT, offsetof(SERVENT_Obj, s_name),    RO },
    { "s_aliases", T_OBJECT, offsetof(SERVENT_Obj, s_aliases), RO },
    { "s_port",    T_SHORT,  offsetof(SERVENT_Obj, s_port),    RO },
    { "s_proto",   T_OBJECT, offsetof(SERVENT_Obj, s_proto),   RO },
    { NULL }
};

static PyObject *SERVENT_getattr(SERVENT_Obj *self, char *name)
{
    return PyMember_Get((char *)self, SERVENT_memberlist, name);
}

static int SERVENT_setattr(SERVENT_Obj *self, char *name, PyObject *v)
{
    return PyMember_Set((char *)self, SERVENT_memberlist, name, v);
}

static char SERVENT__doc__[] = "Wrapper for SERVENT structure";

PyTypeObject SERVENT_Type = {
    PyObject_HEAD_INIT(0)
    0,                          /*ob_size*/
    "SERVENT",                  /*tp_name*/
    sizeof(SERVENT_Obj),        /*tp_basicsize*/
    0,                          /*tp_itemsize*/
    /* methods */
    (destructor)SERVENT_dealloc,/*tp_dealloc*/
    (printfunc)0,               /*tp_print*/
    (getattrfunc)SERVENT_getattr, /*tp_getattr*/
    (setattrfunc)SERVENT_setattr, /*tp_setattr*/
    (cmpfunc)0,                 /*tp_compare*/
    (reprfunc)0,                /*tp_repr*/
    0,                          /*tp_as_number*/
    0,                          /*tp_as_sequence*/
    0,                          /*tp_as_mapping*/
    (hashfunc)0,                /*tp_hash*/
    (ternaryfunc)0,             /*tp_call*/
    (reprfunc)0,                /*tp_str*/

    /* Space for future expansion */
    0L, 0L, 0L, 0L,
    SERVENT__doc__                /* Documentation string */
};

PyObject *SERVENT_new(SERVENT *ent)
{
    SERVENT_Obj *self;
    PyObject *obj;
    int num;

    self = PyObject_NEW(SERVENT_Obj, &SERVENT_Type);
    if (self == NULL)
        return NULL;

    obj = NULL;
    self->s_name = self->s_aliases = self->s_proto = NULL;

    self->s_name = PyString_FromString(ent->s_name);
    if (self->s_name == NULL)
	goto error;

    for (num = 0; ent->s_aliases[num] != NULL; num++)
	;
    self->s_aliases = PyTuple_New(num);
    if (self->s_aliases == NULL)
	goto error;
    for (num = 0; ent->s_aliases[num] != NULL; num++) {
	obj = PyString_FromString(ent->s_aliases[num]);
	if (obj == NULL)
	    goto error;
	if (PyTuple_SetItem(self->s_aliases, num, obj))
	    goto error;
	obj = NULL;		/* stolen */
    }

    self->s_port = ent->s_port;

    self->s_proto = PyString_FromString(ent->s_proto);
    if (self->s_proto == NULL)
	goto error;

    return (PyObject *)self;

error:
    Py_XDECREF(self);
    Py_XDECREF(obj);
    return NULL;
}

PyObject *servent_from_buffer(Buffer_Obj *buff)
{
    if (buff->size <= sizeof(SERVENT)) {
	PyErr_SetString(PyExc_ValueError, "buffer too small");
	return NULL;
    }

    return SERVENT_new((SERVENT *)buff->buff);
}
